//
// Created by wk on 2020/9/17.
//

#include "Processing_message.h"
Processing_message::Processing_message()
{
	m_connect_statu=CONNECT_ERROR;
	mp_socket_receive_data_thread=NULL;
	mp_ros_public_thread=NULL;
	mp_socket_send_data_thread=NULL;

}
Processing_message::~Processing_message()
{
	Processing_message_uninit();
}
void Processing_message::Processing_message_init(std::string server_ip)
{
	m_subscriber=m_nodehandele.subscribe("/move_base_simple/goal",1,goalCallback);//订阅话题
	Processing_message_connect(server_ip);
}
void Processing_message::Processing_message_connect(std::string server_ip)
{
	m_socket.connect(server_ip);//连接目标地址
	Processing_message_run();
}
void Processing_message::Processing_message_band(std::string ip)
{
	m_socket.bind(ip);
	Processing_message_run();
}
//启动程序
void Processing_message::Processing_message_run()
{


	m_connect_statu = CONNECT_NORMAL;
	//默认循环 内部recv等待
	m_socket_receive_condition.reset(false, false, false);
	mp_socket_receive_data_thread = new std::thread(&Processing_message::socket_receive_data_thread,this);//用于接收消息的线程

	//默认循环
	m_ros_public_condition.reset(false, true, false);
	mp_ros_public_thread = new std::thread(&Processing_message::ros_public_thread, this);

	m_socket_send_data_condition.reset(false, true, false);
	mp_socket_send_data_thread = new std::thread(&Processing_message::socket_send_data_thread, this);

}
//反初始化
void Processing_message::Processing_message_uninit()
{
	m_socket_data_queue.termination_queue();
	m_send_data_queue.termination_queue();
	//杀死线程，强制退出
	if (mp_socket_receive_data_thread)
	{
		m_socket_receive_condition.kill_all();
	}

	if (mp_ros_public_thread)
	{
		m_ros_public_condition.kill_all();
	}
	if (mp_socket_send_data_thread)
	{
		m_socket_send_data_condition.kill_all();
	}
	//回收线程的资源
	if (mp_socket_receive_data_thread)
	{
		mp_socket_receive_data_thread->join();
		delete mp_socket_receive_data_thread;
		mp_socket_receive_data_thread = NULL;
	}
	if (mp_ros_public_thread)
	{
		mp_ros_public_thread->join();
		delete mp_ros_public_thread;
		mp_ros_public_thread = NULL;
	}
	if (mp_socket_send_data_thread)
	{
		mp_socket_send_data_thread->join();
		delete mp_socket_send_data_thread;
		mp_socket_send_data_thread = NULL;
	}
	for (auto iter = m_ros_publisher_map.begin(); iter != m_ros_publisher_map.end(); ++iter)
	{
	    delete iter->second;
	}
	m_ros_publisher_map.clear();
	m_connect_statu = CONNECT_ERROR;
	m_socket.close();
}
//接收发送过来的消息
void Processing_message::socket_receive_data_thread()
{
	while (m_socket_receive_condition.is_alive())
	{
		m_socket_receive_condition.wait_for_ex(std::chrono::microseconds(1));
		if ( m_socket_receive_condition.is_alive() )
		{
			std::this_thread::yield();
			std::string t_receive_string;
			{
				std::unique_lock<std::mutex> lk(m_socket_mutex);
				t_receive_string = m_socket.recv<std::string>(1);
			}
			if(t_receive_string.size()>0)//数据长度大于0则说明接收到了消息
			{
					if(m_socket_data_queue.push(t_receive_string))
					{
						std::cout<<"data push success!"<<std::endl;
					}

			}
			t_receive_string.clear();
		}
	}

}
void Processing_message::ros_public_thread()
{
	//通信解析线程, 负责巡检m_receive_data_list, 并解析和处理消息
	while (m_ros_public_condition.is_alive())
	{
	//	m_analysis_data_condition.wait_for_ex(std::chrono::microseconds(1));
		if ( m_ros_public_condition.is_alive() )
		{
			std::this_thread::yield();
			//如果解析线程被主动唤醒, 那么就表示 收到新的消息
			std::string t_queue_data;
			if(m_socket_data_queue.wait_and_pop(t_queue_data))
			{
				message::Base_msg baseMsg;
				baseMsg.ParseFromString(t_queue_data);//第一次解析  判断数据类型
				std::string topic;
				if (baseMsg.base_info().msg_type() == message::ePointCloud_msg )
				{
					message::PointCloud_msg pointCloudMsg;
					pointCloudMsg.ParseFromString(t_queue_data);
					ros::Publisher* search_results=find_topic(pointCloudMsg.topic());
					if( !search_results )//查找是否已存在此topic 若不存在才可new
					{
						search_results=new ros::Publisher(m_nodehandele.advertise<sensor_msgs::PointCloud>(pointCloudMsg.topic(),0));
						m_ros_publisher_map.insert(std::pair<std::string,ros::Publisher*>(pointCloudMsg.topic(),search_results));
						std::this_thread::sleep_for(std::chrono::seconds(1));
					}
					std::cout<<"绘制pointCloud..."<<std::endl;
					sensor_msgs::PointCloud cloud_msg=create_cloudmsg(pointCloudMsg);
					search_results->publish(cloud_msg);
				}
				else if(baseMsg.base_info().msg_type() == message::ePose2d_msg )
				{

					message::Pose2d_msg pose2DMsg;
					pose2DMsg.ParseFromString(t_queue_data);
					ros::Publisher* search_results=find_topic(pose2DMsg.topic());
					if( !search_results )//查找是否已存在此topic 若不存在才可new
					{
						search_results=new ros::Publisher(m_nodehandele.advertise<geometry_msgs::PoseStamped>(pose2DMsg.topic(),true));
						m_ros_publisher_map.insert(std::pair<std::string,ros::Publisher*>(pose2DMsg.topic(),search_results));
					}
					std::cout<<"绘制pose2d..."<<std::endl;
					geometry_msgs::PoseStamped poseStamped=create_posemsg(pose2DMsg);
					search_results->publish(poseStamped);
				}
			}
		}
	}
	return;
}
void Processing_message::socket_send_data_thread()
{
	while (m_socket_send_data_condition.is_alive())
	{
		if ( m_socket_send_data_condition.is_alive() )
		{
			std::this_thread::yield();
			std::string t_send_data;
			if(m_send_data_queue.wait_and_pop(t_send_data))//没有消息就等待 有消息就取出来发送
			{
				std::unique_lock<std::mutex> lk(m_socket_mutex);
				m_socket.send(t_send_data);
			}
		}
	}
	return;
}
sensor_msgs::PointCloud Processing_message::create_cloudmsg(message::PointCloud_msg cloud_pcl)
{
	sensor_msgs::PointCloud cloud;
	cloud.header.stamp = ros::Time::now();
	cloud.header.frame_id = cloud_pcl.frame_id();
	cloud.points.resize(cloud_pcl.points().size());
	//we'll also add an intensity channel to the cloud
	cloud.channels.resize(1);
	cloud.channels[0].name = "wwwkkk";
	cloud.channels[0].values.resize(cloud_pcl.points().size());

	//generate some fake data for our point cloud
	for(unsigned int i = 0; i < cloud_pcl.points().size(); ++i)
	{
		cloud.points[i].x = cloud_pcl.points(i).x();
		cloud.points[i].y = cloud_pcl.points(i).y();
		cloud.points[i].z = cloud_pcl.points(i).z();
		cloud.channels[0].values[i] = 255;
	}
	return cloud;
}
geometry_msgs::PoseStamped Processing_message::create_posemsg(message::Pose2d_msg pose2DMsg)
{
	geometry_msgs::PoseStamped poseStamped;
	poseStamped.header.frame_id=pose2DMsg.frame_id();

	poseStamped.pose.position.x=pose2DMsg.x();
	poseStamped.pose.position.y=pose2DMsg.y();
	poseStamped.pose.position.z=0;
	poseStamped.pose.orientation.x=0*sin(pose2DMsg.theta()/2);
	poseStamped.pose.orientation.y=0*sin(pose2DMsg.theta()/2);
	poseStamped.pose.orientation.z=1*sin(pose2DMsg.theta()/2);
	poseStamped.pose.orientation.w=cos(pose2DMsg.theta()/2);

	return poseStamped;
}
ros::Publisher* Processing_message::find_topic(std::string topic)
{
	for(auto &v : m_ros_publisher_map)
	{
		if(topic==v.first)
			return v.second;
	}
	return NULL;
}

 void  Processing_message::goalCallback(const geometry_msgs::PoseStamped::ConstPtr& msg)
{
	//填充数据

	double x=msg->pose.position.x;
	double y=msg->pose.position.y;
	double z=msg->pose.position.z;
	double qx=msg->pose.orientation.x;
	double qy=msg->pose.orientation.y;
	double qz=msg->pose.orientation.z;
	double qw=msg->pose.orientation.w;

	std::cout<<"  post goal "<<qx<<","<<qy<<","<<qz<<","<<qw<<std::endl;
	double theta=acos(qw)*2.0;
	theta=(qz>0)?theta:-theta;

	message::Goal_msg goal;
	message::Base_info base_info;
	base_info.set_sender(message::eEmpty);
	base_info.set_receiver(message::eAGV);
	base_info.set_msg_type(message::eGoal_msg);
	goal.mutable_base_info()->CopyFrom(base_info);

	goal.set_x(x);
	goal.set_y(y);
	goal.set_theta(theta);

	Processing_message::get_instance_references().m_send_data_queue.push(goal.SerializeAsString());//存入数据
}
